#include "../include/apsrstack-general.h"

void process_ipv4_packet (struct apsrstack_frame_header *header, void *packet, int opt_size) {
  struct apsrstack_ipv4_pkt_header *ipv4_header;
  unsigned short fragment = 0, version = 0, header_length = 0, 
                 old_sum = 0, calc_sum = 0; 
  
  ipv4_header = (struct apsrstack_ipv4_pkt_header *)(packet+data_link_size+opt_size);
  
  version 	= APSR_IPV4_VERSION(ipv4_header);
  header_length = APSR_IPV4_HEADERLENGTH(ipv4_header);
  fragment 	= apsr_ntohs(ipv4_header->fragment);

  if(version != 4) {
	  /* Only IPv4 in this function */
	  stats.ipv4.drops++; /* Drops++ */
	  return;
  }

  if(header_length < 5) {
	  /* header length < 20 bytes */
	  stats.ipv4.drops++; /* Drops++ */
	  return;
  }

  if(header_length * 4 > header->len) {
	  /* header length < frame length */
	  stats.ipv4.drops++; /* Drops++ */
	  return;
  }
 
  if(header_length > 5) {
	  /* IPv4 Options */
	  stats.ipv4.drops++; /* Drops++ */
	  return;
  }
  
  if(fragment && fragment != APSR_IP_DF) {
	  /* Fragmented packet */
	  stats.ipv4.fragmented++; /* Fragmented++ */
	  stats.ipv4.drops++; /* Remove later if reassembling works!  
			       * Drops++ */
	  return;
  }

  /*
   * Checksum verification.
   */
  old_sum = ipv4_header->sum;
  ipv4_header->sum=0x0000;
  calc_sum=in_cksum((unsigned short *)ipv4_header, header_length*4, 0);
  if(old_sum != calc_sum) {
	  #ifdef APSRSTACK_DEBUG
	  apsr_output("DEBUG: Got IPv4 packet with wrong checksum.\n");
          #endif
	  stats.ipv4.checksum_errors++; /* Checksum Errors++*/
	  return;
  }
	  
 
//  /* Only for gateway function, not for host */
//  if(ipv4_header->time_live == 1) {
//          /* TTL == 1 */
//          stats.ipv4.drops++; /* Drops++ */
//          return;
//  }

  if(ipv4_header->src_ip[0] == ipv4_header->dst_ip[0] &&
     ipv4_header->src_ip[1] == ipv4_header->dst_ip[1] &&
     ipv4_header->src_ip[2] == ipv4_header->dst_ip[2] &&
     ipv4_header->src_ip[3] == ipv4_header->dst_ip[3]) {
	  /* Src IPv4 == Dst IPv4 */
          stats.ipv4.drops++; /* Drops++ */
	  return;
  }

  /* If the packet is for us we can go to higher layer(4) */
  if(ipv4_header->dst_ip[0] == APSRSTACK_IP_1 &&
     ipv4_header->dst_ip[1] == APSRSTACK_IP_2 &&
     ipv4_header->dst_ip[2] == APSRSTACK_IP_3 &&
     ipv4_header->dst_ip[3] == APSRSTACK_IP_4) {
	  switchto_ipv4_proto(header, packet, opt_size+header_length*4, ipv4_header->protocol);
	  return;
  }
}

void switchto_ipv4_proto(struct apsrstack_frame_header *header, void *packet, int opt_size, unsigned short protocol) {
  struct ipv4_pkt_header *ipv4_header;
	
  switch(protocol) {
	    case APSR_PROTO_IP:
	    case APSR_PROTO_IP_IN_IP:
              process_ipv4_packet(header, packet, opt_size);
              stats.ipv4.processed++; /* Counter++ */
   	      break;
            case APSR_PROTO_ICMP:
              process_icmpv4_packet(header, packet, opt_size);
              stats.ipv4.processed++; /* Counter++ */
	      break;
            case APSR_PROTO_UDP:
              process_udp_packet(header, packet, opt_size);
              stats.ipv4.processed++; /* Counter++ */
              break;
	    case APSR_PROTO_TCP:
              process_tcp_packet(header, packet, opt_size);
              stats.ipv4.processed++; /* Counter++ */
  	      break;
	    default:
	      /* RFC 1122 :
	       * A host SHOULD generate Destination Unreachable messages with
	       * code:
	       * 2 (Protocol Unreachable), when the designated transport
               *  protocol is not supported;	      
	       */
              ipv4_header = (struct ipv4_pkt_header *)(packet+data_link_size);
    	      
              send_icmpv4(header,
                          5,
                          APSR_ICMPv4,
                          apsr_ntohs(ipv4_header->ident),
                          ipv4_header->dst_ip,
                          ipv4_header->src_ip,
                          APSRSTACK_IPV4_DEFAULT_TTL,
                          APSR_ICMP_DEST_UNREA,
                          APSR_ICMP_CODE_PROTO_UNRE,
			  NULL,
			  0);
              stats.ipv4.dest_un_proto_un++; /* Destination Un./Proto. Un.++ */
  	      break;
  }
}
